import random
import itertools
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence
from helper import fc_block, Constant

def _calculate_accuracy_predicate(logits, batch_target, max_possible_count=None, topk=1, multi_classifier=False):
    batch_size = batch_target.size(0) / max_possible_count

    _, pred = logits.topk(topk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(batch_target.view(1, -1).expand_as(pred))

    k = 1    
    accuray = correct[:k].view(-1).float()
    accuray = accuray.view(-1, max_possible_count)

    correct_k = (accuray.sum(1)==max_possible_count).sum(0)
    correct_k = correct_k * (100.0 / batch_size)
    
    return correct_k

def _calculate_accuracy(
        action_correct,
        object_correct,
        rel_correct,
        target_correct,
        batch_length,
        info):
    
    action_valid_correct = [sum(action_correct[i, :(l - 1)])
                            for i, l in enumerate(batch_length)]
    object_valid_correct = [sum(object_correct[i, :(l - 1)])
                             for i, l in enumerate(batch_length)]
    rel_valid_correct = [sum(rel_correct[i, :(l - 1)])
                             for i, l in enumerate(batch_length)]
    target_valid_correct = [sum(target_correct[i, :(l - 1)])
                             for i, l in enumerate(batch_length)]

    action_accuracy = sum(action_valid_correct).float() / (sum(batch_length) - 1. * len(batch_length))
    object_accuracy = sum(object_valid_correct).float() / (sum(batch_length) - 1. * len(batch_length))
    rel_accuracy = sum(rel_valid_correct).float() / (sum(batch_length) - 1. * len(batch_length))
    target_accuracy = sum(target_valid_correct).float() / (sum(batch_length) - 1. * len(batch_length))

    info.update({'action_accuracy': action_accuracy.cpu().item()})
    info.update({'object_accuracy': object_accuracy.cpu().item()})
    info.update({'rel_accuracy': rel_accuracy.cpu().item()})
    info.update({'target_accuracy': target_accuracy.cpu().item()})

class PredicateClassifier(nn.Module):

    def __init__(
            self, 
            args,
            dset,
            loss_weight
            ):

        super(PredicateClassifier, self).__init__()

        self.num_goal_predicates = dset.num_goal_predicates
        self.max_possible_count = dset.max_goal_length
        self.loss = custom_loss(loss_weight)

        hidden_size = args.demo_hidden
        self.hidden_size = hidden_size
        self.loss_type = args.loss_type
        
        if args.dropout==0:
            print('dropout', args.dropout)
            classifier = nn.Sequential()
            classifier.add_module('fc_block1', fc_block(hidden_size*82, hidden_size, False, nn.Tanh))
            classifier.add_module('fc_block2', fc_block(hidden_size, 79, False, None)) # 79 is all possible actions
        else:
            print('dropout not 0', args.dropout)
            classifier = nn.Sequential()
            classifier.add_module('fc_block1', fc_block(hidden_size*82, hidden_size, False, nn.Tanh))
            classifier.add_module('dropout', nn.Dropout(args.dropout))
            classifier.add_module('fc_block2', fc_block(hidden_size, 79, False, None)) # 79 is all possible actions

        self.classifier = classifier

    def forward(self, input_emb, batch_target, batch_action_id, batch_task_name, **kwargs):
        input_emb = input_emb.view(-1, self.hidden_size*82)
        logits = self.classifier(input_emb)

        prob = F.softmax(logits, 1)

        #cross entropy loss
        if self.loss_type == 'ce':
            loss = F.cross_entropy(logits, batch_target)  
        #custom loss
        if self.loss_type == 'regu':
            loss = self.loss(logits, batch_target)  

        argmax_Y = torch.max(logits, 1)[1].view(-1, 1)
        top1 = (batch_target.float().view(-1, 1) == argmax_Y.float()).sum().item() / len(batch_target.float().view(-1, 1)) * 100

        with torch.no_grad():
            info = {
                "prob": prob.cpu().numpy(),
                "argmax": argmax_Y.cpu().numpy(),
                "loss": loss.cpu().numpy(),
                "top1": top1,
                "target": batch_target.cpu().numpy(),
                "task_name": batch_task_name,
                "action_id": batch_action_id.cpu().numpy()
            }

        return loss, info

class custom_loss(nn.Module):
	def __init__(self, loss_weight) -> None:
		super(custom_loss, self).__init__( )
		self.loss = nn.CrossEntropyLoss().cuda()
		self.weight_loss = nn.MSELoss().cuda()		
		self.loss_weight = loss_weight
		self.counts = torch.FloatTensor(self.loss_weight).cuda()

	def forward(self, pred, target):		
		# weight loss + cross entropy loss
		batch_counts = torch.bincount(target)
		batch_counts = batch_counts/torch.sum(batch_counts)
		if len(batch_counts) < 79:
			batch_counts = F.pad(input=batch_counts, pad=(0, 79 - len(batch_counts)%79), mode='constant', value=0)
		
		celoss = self.loss(pred, target)
		customloss = self.weight_loss(batch_counts, self.counts)
		print('celoss: ', celoss, 'customloss: ', customloss)
		loss = celoss +  1000*customloss
		return loss

class PredicateClassifierMultiClassifier(nn.Module):

    def __init__(
            self, 
            args,
            dset):

        super(PredicateClassifierMultiClassifier, self).__init__()

        self.num_goal_predicates = dset.num_goal_predicates
        self.max_possible_count = dset.max_goal_length
        self.max_subgoal_length = dset.max_subgoal_length

        hidden_size = args.demo_hidden
        print('hidden_size', hidden_size)
        print('PredicateClassifierMultiClassifier')

        if args.dropout==0:
            print('dropout', args.dropout)
            classifier = nn.Sequential()
            classifier.add_module('fc_block1', fc_block(hidden_size, hidden_size, False, nn.Tanh))
            classifier.add_module('fc_block2', fc_block(hidden_size, self.num_goal_predicates*(self.max_subgoal_length+1), False, None))
        else:
            print('dropout not 0', args.dropout)
            classifier = nn.Sequential()
            classifier.add_module('fc_block1', fc_block(hidden_size, hidden_size, False, nn.Tanh))
            classifier.add_module('dropout', nn.Dropout(args.dropout))
            classifier.add_module('fc_block2', fc_block(hidden_size, self.num_goal_predicates*(self.max_subgoal_length+1), False, None))

        
        self.classifier = classifier

    def forward(self, bs, input_emb, batch_target, batch_file_name, **kwargs):
        
        logits = self.classifier(input_emb)
        logits = logits.reshape([-1, (self.max_subgoal_length+1)])
        prob = F.softmax(logits, 1)

        batch_target = torch.cat(batch_target)
        loss = F.cross_entropy(logits, batch_target)  

        top1 = _calculate_accuracy_predicate(logits, batch_target, self.num_goal_predicates, multi_classifier=True)

        with torch.no_grad():
            info = {
                "prob": prob.cpu().numpy(),
                "loss": loss.cpu().numpy(),
                "top1": top1.cpu().numpy(),
                "target": batch_target.cpu().numpy(),
                "file_name": batch_file_name
            }

        return loss, info

class ActionDemoEncoder(nn.Module):
    def __init__(self, args, dset, pooling):
        super(ActionDemoEncoder, self).__init__()
        hidden_size = args.demo_hidden
        self.hidden_size = hidden_size

        len_action_predicates = dset.max_action_len
        self.action_embed = nn.Embedding(len_action_predicates, hidden_size)

        feat2hidden = nn.Sequential()
        feat2hidden.add_module(
            'fc_block1', fc_block(hidden_size, hidden_size, False, nn.ReLU))
        self.feat2hidden = feat2hidden

        self.pooling = pooling

        if 'lstm' in self.pooling:
            self.lstm = nn.LSTM(hidden_size, hidden_size)


    def forward(self, batch_data, batch_gt, batch_task_name):        
        batch_data = batch_data.view(-1,1)

        stacked_demo_feat = self.action_embed(batch_data)

        stacked_demo_feat = self.feat2hidden(stacked_demo_feat)
        batch_demo_feat = []
        start = 0
        
        for length in range(0,batch_data.shape[0]):
            if length == 0:
                feat = stacked_demo_feat[0:1, :]
            else:
                feat = stacked_demo_feat[(length-1):length, :]
            if len(feat.size()) == 3:
                feat = feat.unsqueeze(0)
            
            if self.pooling == 'max':
                feat = torch.max(feat, 0)[0]
            elif self.pooling == 'avg':
                feat = torch.mean(feat, 0)
            elif self.pooling == 'lstmavg':
                lstm_out, hidden = self.lstm(feat.view(len(feat), 1, -1))
                lstm_out = lstm_out.view(len(feat), -1)
                feat = torch.mean(lstm_out, 0)
            elif self.pooling == 'lstmlast':
                lstm_out, hidden = self.lstm(feat.view(len(feat), 1, -1))
                lstm_out = lstm_out.view(len(feat), -1)
                feat = lstm_out[-1]
            else:
                raise ValueError


            batch_demo_feat.append(feat)

        demo_emb = torch.stack(batch_demo_feat, 0)
        demo_emb = demo_emb.view(8,82, -1)
        return demo_emb, batch_demo_feat

